/*
 * Routines to wrap lines and indent them.
 *
 * Bert Bos
 * Created 10 May 1998
 * $Id: textwrap.c,v 1.7 1999/05/25 15:52:11 bbos Exp $
 */
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include "export.h"
#include "types.e"
/*
#include "tree.e"
#include "html.tab.e"
#include "scan.e"
#include "api.h"
*/

#define MAXLINE 4096
#define NBSP 128				/* Marks non-break-space */

static unsigned char buf[MAXLINE];
static int len = 0;
static int linelen = 0;
static int level = 0;
static int indent = 2;
static int maxlinelen = 72;

/* set_indent -- set the amount of indent per level */
EXPORT void set_indent(int n) {indent = n;}

/* set_linelen -- set the maximum length of a line */
EXPORT void set_linelen(int n) {maxlinelen = n;}

/* flush -- print word in buf */
EXPORT void flush()
{
  int i, j;

  while (len != 0 && linelen + len >= maxlinelen) { /* Line needs break */
    /* Find last space before maxlinelen */
    for (i = maxlinelen - linelen - 1; i >= 0 && buf[i] != ' '; i--) ;
    if (i >= 0) {				/* Found a breakpoint at i */
      /* Print up to breakpoint (removing non-break-space markers) */
      for (j = 0; j < i; j++) putchar(buf[j] != NBSP ? buf[j] : ' ');
      putchar('\n');				/* Break line */
      linelen = 0;
      memmove(buf + level * indent, buf + i + 1, len - i - 1); /* Indent */
      for (j = 0; j < level * indent; j++) buf[j] = NBSP;
      len += level * indent - i - 1;
    } else {					/* No breakpoint found */
      /* Print whole buf (removing non-break-space markers) */
      for (j = 0; j < len; j++) putchar(buf[j] != NBSP ? buf[j] : ' ');
      linelen += len;
      len = 0;
    }
  }
  /* Print rest, if any (removing non-break-space markers) */
  for (j = 0; j < len; j++) putchar(buf[j] != NBSP ? buf[j] : ' ');
  linelen += len;
  len = 0;
}

/* outc -- add one character to output buffer */
EXPORT void outc(unsigned char c, Boolean preformatted)
{
  if (c == '\n' && !preformatted) c = ' ';	/* Newline is just a space */
  if (c == '\r' && !preformatted) c = ' ';	/* CR is just a space */
  if (c == '\t' && !preformatted) c = ' ';	/* Tab is just a space */
  if (c == '\f' && !preformatted) c = ' ';	/* Formfeed is just a space */
  if (c == ' ' && preformatted) c = NBSP;	/* Non-break-space marker */
  if (c == ' ' && linelen + len >= maxlinelen) flush();	/* Empty the buf */
  if (c == ' ' && linelen + len == 0) return;	/* Don't insert space at BOL */
  if (linelen + len == 0) while (len < level * indent) buf[len++] = NBSP;
  if (c == ' ' && len && buf[len-1] == ' ') return; /* Skip multiple spaces */
  buf[len++] = c;				/* Finally, insert c */
}

/* out -- add text to current output line, print line if getting too long */
EXPORT void out(unsigned char *s, Boolean preformatted)
{
  if (s) for (; *s; s++) outc(*s, preformatted);
}

/* outn -- add n chars to current output, print line if getting too long */
EXPORT void outn(unsigned char *s, size_t n, Boolean preformatted)
{
  size_t i;
  for (i = 0; i < n; i++) outc(s[i], preformatted);
}

/* outln -- add string to output buffer, followed by '\n' */
EXPORT void outln(unsigned char *s, Boolean preformatted)
{
  out(s, preformatted);
  flush();
  assert(len == 0);
  putchar('\n');
  linelen = 0;
}

/* outbreak -- conditional new line; make sure next text starts on new line */
EXPORT void outbreak()
{
  flush();
  assert(len == 0);
  if (linelen != 0) {
    putchar('\n');
    linelen = 0;
  }
}

/* inc_indent -- increase indentation level by 1 */
EXPORT void inc_indent(void)
{
  flush();
  level++;
}

/* decc_indent -- decrease indentation level by 1 */
EXPORT void dec_indent(void)
{
  flush();
  level--;
}
