/*
 * List all links from the given document(s).
 * Links are found as HREF, SRC and DATA attributes.
 * Option -l gives a long listing, including tag name and REL attribute.
 * Option -v gives a multi-line listing, including ALT and TITLE.
 *
 * Bert Bos <bert@w3.org>
 * Created 31 July 1999
 * $Id: wls.c,v 1.2 1999/07/31 19:46:49 bbos Exp $
 */
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "export.h"
#include "types.e"
#include "tree.e"
#include "html.e"
#include "scan.e"
#include "openurl.e"
#include "heap.e"
#include "url.e"

static string base;				/* Current URL */
static int verbosity = 0;			/* How much info to print */


/* down -- convert a string to lowercase, return pointer to arg */
static inline string down(string s)
{
  string t;
  for (t = s; *t; t++) *t = tolower(*t);
  return s;
}

/* attval -- get value of a specific attribute, or NULL */
static string  attval(pairlist attribs, const string name)
{
  pairlist p;

  for (p = attribs; p; p = p->next)
    if (strcasecmp(p->name, name) == 0) return p->value ? p->value : p->name;
  return NULL;
}

/* handle_error -- called when a parse error occurred */
EXPORT void handle_error(void *clientdata, const string s, int lineno)
{
  fprintf(stderr, "%d: %s\n", lineno, s);
}

/* start -- called before the first event is reported */
EXPORT void* start(void) {return NULL;}
  
/* end -- called after the last event is reported */
EXPORT void end(void *clientdata) {}

/* handle_comment -- called after a comment is parsed */
EXPORT void handle_comment(void *clientdata, string commenttext)
{
  dispose(commenttext);
}

/* handle_text -- called after a text chunk is parsed */
EXPORT void handle_text(void *clientdata, string text)
{
  dispose(text);
}

/* handle_declaration -- called after a declaration is parsed */
EXPORT void handle_decl(void *clientdata, string gi,
			string fpi, string url)
{
  dispose(gi);
  dispose(fpi);
  dispose(url);
}

/* handle_proc_instr -- called after a PI is parsed */
EXPORT void handle_pi(void *clientdata, string pi_text)
{
  dispose(pi_text);
}

/* handle_tag -- called after a start tag or an empty tag is parsed */
static void handle_tag(void *clientdata, string name, pairlist attribs)
{
  string uri, rel, full, title, alt;

  if (!strcasecmp(name, "base") && (uri = attval(attribs, "href"))) {
    /* Handle <BASE> specially: new base URI */
    dispose(base);
    base = newstring(uri);
  } else {
    /* Anything that has HREF, SRC or DATA is considered a link */
    if ((uri = attval(attribs, "href"))
	|| (uri = attval(attribs, "src"))
	|| (uri = attval(attribs, "data"))) {
      full = URL_s_absolutize(base, uri);
      switch (verbosity) {
	case 0:					/* Short listing */
	  printf("%s\n", full);
	  break;
	case 1:					/* Long listing */
	  rel = attval(attribs, "rel");
	  printf("%-7s %-10s %s\n",down(name),rel?down(rel):(string)"",full);
	  break;
	case 2:					/* Verbose listing */
	  rel = attval(attribs, "rel");
	  title = attval(attribs, "title");
	  alt = attval(attribs, "alt");
	  printf("URI: %s\n", full);
	  printf("element: %s\n", down(name));
	  printf("rel: %s\n", rel ? down(rel) : (string) "");
	  printf("title: %s\n", title ? title : (string) "");
	  printf("alt: %s\n\n", alt ? alt : (string) "");
	  break;
	default:
	  assert(!"Cannot happen");
      }
    }
  }
  pairlist_delete(attribs);
  dispose(name);
}

/* handle_starttag -- called after a start tag is parsed */
EXPORT void handle_starttag(void *clientdata, string name, pairlist attribs)
{
  handle_tag(clientdata, name, attribs);
}

/* handle_emptytag -- called after an empty tag is parsed */
EXPORT void handle_emptytag(void *clientdata, string name, pairlist attribs)
{
  handle_tag(clientdata, name, attribs);
}

/* handle_endtag -- called after an endtag is parsed (name may be "") */
EXPORT void handle_endtag(void *clientdata, string name)
{
  dispose(name);
}

/* usage -- print usage message and exit */
static void usage(string name)
{
  fprintf(stderr,
	  "Usage: %s [ -l | -v ] [ html-file [ html-file... ] ]\n", name);
  exit(1);
}

int main(int argc, char *argv[])
{
  int i;

  /* Bind the parser callback routines to our handlers */
  set_error_handler(handle_error);
  set_start_handler(start);
  set_end_handler(end);
  set_comment_handler(handle_comment);
  set_text_handler(handle_text);
  set_decl_handler(handle_decl);
  set_pi_handler(handle_pi);
  set_starttag_handler(handle_starttag);
  set_emptytag_handler(handle_emptytag);
  set_endtag_handler(handle_endtag);

  /* Check command line options */
  for (i = 1; argc > i && argv[i][0] == '-'; i++) {
    if (eq(argv[i], "-l")) verbosity = 1;
    else if (eq(argv[i], "-v")) verbosity = 2;
    else if (eq(argv[i], "--")) break;
    else usage(argv[0]);
  }

  /* Loop over files/urls */
  if (argc == i) {				/* No file, use stdin */
    base = newstring("");
    yyin = stdin;
    if (yyparse() != 0) exit(3);
  } else {
    for (; argc > i; i++) {
      base = newstring(argv[i]);
      yyin = fopenurl(argv[i], "r");
      if (yyin == NULL) {perror(argv[i]); exit(2);}
      if (yyparse() != 0) exit(3);
      fclose(yyin);
    }
  }

  return 0;
}
